from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum

from asgiref.sync import async_to_sync
import opendal.exceptions as opendal_exceptions


class StorageScheme(Enum):
    S3 = "S3"
    AZURE_BLOB = "AzureBlob"
    DROPBOX = "Dropbox"


@dataclass
class JSONStorageDescriptor:
    @staticmethod
    def create_from_json_descriptor(json_descriptor):
        from .s3.descriptor import S3JSONStorageDescriptor

        if json_descriptor["scheme"] == StorageScheme.S3.value:
            return S3JSONStorageDescriptor(
                bucket=json_descriptor["bucket"],
                key=json_descriptor["key"],
                aws_access_key_id=json_descriptor["awsAccessKeyId"],
                encrypted_aws_secret_access_key=json_descriptor[
                    "encryptedAwsSecretAccessKey"
                ],
                region=json_descriptor.get("region"),
                endpoint=json_descriptor.get("endpoint"),
            )
        else:
            raise ValueError(f"Unsupported storage scheme: {json_descriptor['scheme']}")


@dataclass
class StorageDescriptor(ABC):
    path: str

    @abstractmethod
    def decrypt(self):
        self.is_decrypted = True

    @staticmethod
    def create_from_json_storage_descriptor(
        json_descriptor: JSONStorageDescriptor,
    ):
        from .s3.descriptor import S3JSONStorageDescriptor, S3StorageDescriptor

        if isinstance(json_descriptor, S3JSONStorageDescriptor):
            return S3StorageDescriptor(
                bucket=json_descriptor.bucket,
                path=json_descriptor.key,
                aws_access_key_id=json_descriptor.aws_access_key_id,
                encrypted_aws_secret_access_key=json_descriptor.encrypted_aws_secret_access_key,
                region=json_descriptor.region,
                endpoint=json_descriptor.endpoint,
            )
        else:
            raise ValueError(f"Unsupported storage scheme: {json_descriptor.scheme}")


class OpenDALClient(ABC):
    @staticmethod
    def create_from_storage_descriptor(
        storage_descriptor: StorageDescriptor,
    ):
        from .s3.client import OpenDALS3Client
        from .s3.descriptor import S3StorageDescriptor

        if not storage_descriptor.is_decrypted:
            raise ValueError(
                "Storage descriptor must be decrypted before creating OpenDAL client."
            )
        if isinstance(storage_descriptor, S3StorageDescriptor):
            return OpenDALS3Client(
                bucket=storage_descriptor.bucket,
                region=storage_descriptor.region,
                endpoint=storage_descriptor.endpoint,
                access_key_id=storage_descriptor.aws_access_key_id,
                secret_access_key=storage_descriptor.aws_secret_access_key,
            )
        else:
            raise ValueError(f"Unsupported storage type: {storage_descriptor.scheme}")

    def read(self, path: str):
        return self.operator.read(path)
    
    def presign_read(self, path: str, expiration: int):
        async def get_presigned_url():
            presigned_read = await self.operator.to_async_operator().presign_read(path, expiration)
            return presigned_read
        return async_to_sync(get_presigned_url)()

    def stat(self, path: str):
        return self.operator.stat(path)

    def open(self, path: str, mode: str):
        return self.operator.open(path, mode)

    def write(self, path: str, data: bytes):
        self.operator.write(path, data)

    def delete(self, path: str):
        try:
            self.operator.delete(path)
            self.stat(path)
        except opendal_exceptions.NotFound:
            # need to explicity catch the failed read as OpenDAL doesn't error when delete fails
            pass
        except Exception as e:
            print(e)
            assert False, f"{path} not deleted"




import logging
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory

import boto3
from boto3.s3.transfer import TransferConfig
from botocore.client import Config
from botocore.exceptions import ClientError
from tqdm import tqdm

from ..utils import run_shell_command


class S3Client:
    def __init__(
        self, profile_name=None, credentials=None, endpoint_url=None, region_name=None
    ):
        """
        Initialize the S3Client with optional AWS credentials and configuration.

        :param profile_name: AWS CLI profile name to use.
        :param credentials: Dictionary containing 'aws_access_key_id' and 'aws_secret_access_key'.
        :param endpoint_url: Custom S3 endpoint URL.
        :param region_name: AWS region name.
        """
        session_kwargs = {}
        client_kwargs = {}

        # Configure session with profile if provided
        if profile_name:
            session_kwargs["profile_name"] = profile_name

        # Initialize boto3 session
        self.session = boto3.session.Session(**session_kwargs)

        # Handle credentials dictionary
        if credentials:
            assert "aws_access_key_id" in credentials
            assert "aws_secret_access_key" in credentials
            client_kwargs["aws_access_key_id"] = credentials["aws_access_key_id"]
            client_kwargs["aws_secret_access_key"] = credentials[
                "aws_secret_access_key"
            ]

        # Configure other client parameters
        if endpoint_url:
            client_kwargs["endpoint_url"] = endpoint_url
        if region_name:
            client_kwargs["region_name"] = region_name

        # Set S3 client configuration
        client_kwargs["config"] = Config(signature_version="s3v4")

        # Initialize S3 client
        self.client = self.session.client("s3", **client_kwargs)

        # Store configuration for reference
        self.profile_name = profile_name
        self.credentials = credentials
        self.endpoint_url = endpoint_url
        self.region_name = region_name

    @contextmanager
    def download_to_temporary_file(
        self, bucket, key, file_name=None, show_progress=False
    ):
        try:
            temp_directory = TemporaryDirectory()
            if file_name is None:
                file_name = key.split("/")[-1]
            save_path = Path(temp_directory.name) / file_name
            self.download(bucket, key, str(save_path), show_progress)
            yield save_path
        finally:
            temp_directory.cleanup()

    def download(
        self,
        bucket,
        key,
        save_prefix,
        show_progress=False,
        Config=TransferConfig(),
    ):
        save_prefix = str(save_prefix)

        if show_progress:

            def hook(t):
                def inner(bytes_amount):
                    t.update(bytes_amount)

                return inner

            file_size = float(
                self.client.head_object(Bucket=bucket, Key=key)["ContentLength"]
            )
            with tqdm(
                total=file_size, unit="B", unit_scale=True, desc=save_prefix
            ) as t:
                self.client.download_file(
                    bucket, key, save_prefix, Callback=hook(t), Config=Config
                )
        else:
            self.client.download_file(bucket, key, save_prefix, Config=Config)

    def upload_file(self, path, bucket, key, Config=TransferConfig(), **kwargs):
        self.client.upload_file(path, bucket, key, Config=Config, **kwargs)

    def upload_fileobj(self, fileobj, bucket, key, Config=TransferConfig(), **kwargs):
        self.client.upload_fileobj(fileobj, bucket, key, Config=Config, **kwargs)

    def copy(self, source_bucket, source_key, dst_bucket, dst_key):
        self.client.copy(
            {"Bucket": source_bucket, "Key": source_key}, dst_bucket, dst_key
        )

    def check_if_exists_in_s3(self, bucket, key):
        try:
            self.client.head_object(Bucket=bucket, Key=key)
            return True
        except Exception:
            return False

    def get_object_type(self, bucket, key):
        return self.client.head_object(Bucket=bucket, Key=key)["ContentType"]

    def get_object_size(self, bucket, key):
        return self.client.head_object(Bucket=bucket, Key=key)["ContentLength"]

    def head_object(self, bucket, key):
        return self.client.head_object(Bucket=bucket, Key=key)

    def get_object_storage_class(self, bucket, key):
        return self.client.head_object(Bucket=bucket, Key=key).get(
            "StorageClass", "STANDARD"
        )

    def delete_directory(self, bucket, prefix):
        keys = self.get_s3_keys(bucket, prefix)
        for key in keys:
            self.delete_object(bucket, key)

    def delete_object(self, bucket, key):
        self.client.delete_object(Bucket=bucket, Key=key)

    def get_s3_keys(self, bucket, prefix=None, delimiter=None):
        keys = []
        try:
            for resp in self._construct_s3_paginator(bucket, prefix, delimiter):
                keys.extend([c["Key"] for c in resp["Contents"]])
        except KeyError:
            pass
        return keys

    def get_top_level_bucket_keys(self, bucket):
        keys = []
        for prefix in self._construct_s3_paginator(bucket, delimiter="/").search(
            "CommonPrefixes"
        ):
            keys.append(prefix.get("Prefix"))
        return keys

    def get_directories_for_bucket_with_prefix_recursively(
        self, bucket, prefix=None, delimiter="/", log_every=100
    ):
        prefixes_parsed = 0
        stack = [prefix]
        all_prefixes = []
        while len(stack) > 0:
            prefix = stack.pop()
            if prefix is None:
                sub_prefixes = self.get_directories_for_bucket_with_prefix(
                    bucket, delimiter=delimiter
                )
            else:
                all_prefixes.append(prefix)
                sub_prefixes = self.get_directories_for_bucket_with_prefix(
                    bucket, prefix, delimiter
                )
            stack.extend(sub_prefixes)
            prefixes_parsed += 1
            if prefixes_parsed % log_every == 0:
                print(f"Parsed: {prefixes_parsed}")
        return all_prefixes

    def get_directories_for_bucket_with_prefix(
        self, bucket, prefix=None, delimiter="/"
    ):
        keys = []
        for prefix in self._construct_s3_paginator(
            bucket, prefix=prefix, delimiter=delimiter
        ).search("CommonPrefixes"):
            if prefix is None:
                continue
            keys.append(prefix.get("Prefix"))
        return keys

    def get_files_for_bucket_with_prefix(self, bucket, prefix, delimiter="/"):
        all_file_contents = []
        for contents in self._construct_s3_paginator(
            bucket, prefix=prefix, delimiter=delimiter
        ).search("Contents"):
            if contents is None:
                continue
            if contents["Key"] == prefix:
                continue
            all_file_contents.append(contents)
        return all_file_contents

    def generate_presigned_get_url(self, bucket, key, expiration=3600):
        """Generate a presigned URL to share an S3 object
        :param bucket_name: string
        :param object_name: string
        :param expiration: Time in seconds for the presigned URL to remain
        valid
        :return: Presigned URL as string. If error, returns None.
        """

        # Generate a presigned URL for the S3 object
        try:
            response = self.client.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket, "Key": key},
                ExpiresIn=expiration,
            )
        except ClientError as e:
            logging.error(e)
            return None

        # The response contains the presigned URL
        return response

    def create_presigned_post_url(
        self, bucket, key, fields=None, conditions=None, expiration=3600
    ):
        """Generate a presigned URL S3 POST request to upload a file

        :param bucket_name: string
        :param object_name: string
        :param fields: Dictionary of prefilled form fields
        :param conditions: List of conditions to include in the policy
        :param expiration: Time in seconds for the presigned URL to remain
        valid
        :return: Dictionary with the following keys:
            url: URL to post to
            fields: Dictionary of form fields and values to submit with the
            POST
        :return: None if error.
        """

        # Generate a presigned S3 POST URL
        try:
            response = self.client.generate_presigned_post(
                bucket,
                key,
                Fields=fields,
                Conditions=conditions,
                ExpiresIn=expiration,
            )
        except ClientError as e:
            logging.error(e)
            return None

        # The response contains the presigned URL and required fields
        return response

    def sync_folder(
        self,
        source_bucket,
        source_prefix,
        destination_bucket,
        destination_prefix,
    ):
        run_shell_command(
            " ".join(
                [
                    "aws",
                    "s3",
                    "sync",
                    f'"s3://{source_bucket}/{source_prefix}"',
                    f'"s3://{destination_bucket}/{destination_prefix}"',  # noqa
                ]
            )
        )

    def sync_folder_from_local_to_s3(
        self,
        local_folder_path,
        destination_bucket,
        destination_prefix,
    ):
        run_shell_command(
            " ".join(
                [
                    "aws",
                    "s3",
                    "sync",
                    local_folder_path,
                    f'"s3://{destination_bucket}/{destination_prefix}"',
                ]
            )
        )

    def upload_directory(self, directory_path, bucket, prefix):
        # Convert the local path to a Path object
        directory_path = Path(directory_path)

        # Walk through the local directory and upload each file
        for child in directory_path.rglob("*"):
            # Check if the child is a file
            if child.is_file():
                # Compute the full S3 key of the file
                s3_key = str(Path(prefix) / child.relative_to(directory_path))

                # Upload the file to S3
                self.upload_file(str(child), bucket, s3_key)

    def _construct_s3_paginator(self, bucket, prefix=None, delimiter=None):
        kwargs = {"Bucket": bucket}
        s3_paginator = self.client.get_paginator("list_objects_v2")

        if prefix is not None:
            kwargs.update({"Prefix": prefix})

        if delimiter is not None:
            kwargs.update({"Delimiter": delimiter})
        paginate = s3_paginator.paginate(**kwargs)
        return paginate



from pathlib import Path
from tempfile import TemporaryDirectory
from contextlib import contextmanager

from opendal import Operator


class OpenDALService:
    def copy(
        self,
        source_client,
        source_path,
        destination_client,
        destination_path,
        read_chunk_size=1024,
    ):
        with source_client.open(source_path, "rb") as source_file:
            with destination_client.open(destination_path, "wb") as destination_file:
                bytes_to_write = source_client.stat(source_path).content_length
                while bytes_to_write > 0:
                    if bytes_to_write < read_chunk_size:
                        chunk_size_to_read = bytes_to_write
                    else:
                        chunk_size_to_read = read_chunk_size
                    chunk = bytes(source_file.read(chunk_size_to_read))
                    destination_file.write(chunk)
                    bytes_to_write -= chunk_size_to_read

    def download(self, download_client, download_from_path, save_path):
        source_client = download_client
        source_path = download_from_path
        destination_client = Operator("fs", root="/")
        destination_path = str(save_path)
        self.copy(source_client, source_path, destination_client, destination_path)

    @contextmanager
    def download_to_temporary_file(self, download_client, download_from_path, file_name=None):
        download_from_path = str(download_from_path)
        try:
            temp_directory = TemporaryDirectory()
            if file_name is None:
                file_name = download_from_path.split("/")[-1]
            save_path = Path(temp_directory.name) / file_name
            self.download(download_client, download_from_path, str(save_path))
            yield save_path
        finally:
            temp_directory.cleanup()



from opendal import Operator

from ..shared import OpenDALClient


class OpenDALS3Client(OpenDALClient):
    def __init__(
        self,
        bucket,
        region=None,
        endpoint=None,
        access_key_id=None,
        secret_access_key=None,
    ):
        """
        Initialize the OpenDALS3Client with AWS credentials and configuration.

        :param bucket: Name of the S3 bucket.
        :param region: AWS region name.
        :param endpoint: Custom S3 endpoint URL.
        :param access_key: AWS access key ID.
        :param secret_key: AWS secret access key.
        """
        s3_config = {
            "bucket": bucket,
        }

        if region:
            s3_config["region"] = region
        if endpoint:
            s3_config["endpoint"] = endpoint
        if access_key_id:
            s3_config["access_key_id"] = access_key_id
        if access_key_id:
            s3_config["secret_access_key"] = secret_access_key

        # Initialize opendal S3 operator
        self.operator = Operator("s3", **s3_config)
        self._bucket = bucket
        self._region = region
        self._endpoint = endpoint
        self._access_key_id = access_key_id
        self._secret_access_key = secret_access_key

    @property
    def bucket(self):
        return self._bucket

    @property
    def region(self):
        return self._region

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def access_key_id(self):
        return self._access_key_id

    @property
    def secret_access_key(self):
        return self._secret_access_key



import subprocess
import sys


def run_shell_command(cmd, silent=False, same_process=False):
    try:
        if not silent:
            print(cmd)
        if same_process:
            process = subprocess.Popen(
                ["bash"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
            )
            process.stdin.write(cmd.encode(sys.stdout.encoding))
            output, _ = process.communicate()
        else:
            process = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if not silent:
                for line in iter(process.stdout.readline, b""):
                    sys.stdout.write(line.decode(sys.stdout.encoding))
    except subprocess.CalledProcessError as e:
        print(e.output.decode())



from setuptools import find_packages, setup

import pys3thon

setup(
    name="pys3thon",
    version=pys3thon.__version__,
    author="Kevin Lu",
    author_email="kevinyihchyunlu@gmail.com",
    license="MIT",
    packages=find_packages(exclude=["tests", "tests/*"]),
    install_requires=[
        "boto3>=1.24.89",
        "tqdm==4.64.0",
        "pyOpenSSL==22.1.0",
        "opendal>=0.45.2",
        "asgiref==3.8.1"
    ],
    extras_require={
        "dev": [
            "pytest==7.1.3",
            "pytest-mock==3.10.0",
            "pytest-only",
            "python-semantic-release==7.32.1",
            "moto>=5.0",
            "flake8==5.0.4",
            "black==22.10.0",
            "isort==5.10.1",
            "pytest-watch==4.2.0",
            "pre-commit==2.20.0",
            "click==8.1.3",
            "python-dotenv==1.0.1",
        ]
    },
)



from dataclasses import dataclass
from typing import Optional

from ..shared import JSONStorageDescriptor, StorageDescriptor


@dataclass
class S3JSONStorageDescriptor(JSONStorageDescriptor):
    bucket: str
    key: str
    aws_access_key_id: str
    encrypted_aws_secret_access_key: str
    region: Optional[str] = None
    endpoint: Optional[str] = None


@dataclass
class S3StorageDescriptor(StorageDescriptor):
    bucket: str
    aws_access_key_id: str
    encrypted_aws_secret_access_key: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    region: Optional[str] = None  # can be set via environment variable
    endpoint: Optional[str] = None

    is_decrypted: bool = False

    def __post_init__(self):
        assert (
            self.encrypted_aws_secret_access_key is not None
            or self.aws_secret_access_key is not None
        )

        if self.aws_secret_access_key is not None:
            self.is_decrypted = True

    def decrypt(self, decrypt_fn):
        if self.is_decrypted:
            return
        super().decrypt()
        self.aws_secret_access_key = decrypt_fn(self.encrypted_aws_secret_access_key)
